use std::{collections::HashMap, ops::Range, process};

use codespan_reporting::{
    diagnostic::{Diagnostic, Label, Severity},
    files::SimpleFiles,
    term::{
        self,
        termcolor::{ColorChoice, StandardStream},
    },
};

pub type Span = Range<usize>;

type FileId = usize;

#[derive(PartialEq, Eq, Hash, Copy, Clone)]
pub struct FileIdent<'a> {
    id: FileId,
    name: &'a str,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum DiagnosticKind {
    InvalidToken,
}

pub enum DiagCriticality {
    /// Bugs in this program, unreachable code being reached, etc.
    Bug,

    // TODO: Better names
    /// Critical errors that cannot be recovered from, prompt immediate exit
    CriticalImmediate,
    /// Critical, this program must exit after this processing step, but can continue and may find other diagnostics
    CriticalFinishStep,

    /// A warning, about to be deprecated syntax etc.
    Warn,

    /// Information/advice
    Info,
    Help,
}

impl DiagCriticality {
    fn mk_diag_base(&self) -> Diagnostic<FileId> {
        Diagnostic::new(match self {
            DiagCriticality::Bug => Severity::Bug,
            DiagCriticality::CriticalImmediate => Severity::Error,
            DiagCriticality::CriticalFinishStep => Severity::Error,
            DiagCriticality::Warn => Severity::Warning,
            DiagCriticality::Info => Severity::Note,
            DiagCriticality::Help => Severity::Help,
        })
    }
}

impl DiagnosticKind {
    pub fn criticality(&self) -> DiagCriticality {
        match self {
            DiagnosticKind::InvalidToken => DiagCriticality::CriticalFinishStep,
        }
    }

    fn into_diag(self, locs: &[Span], file_id: &FileIdent) -> Diagnostic<FileId> {
        self.criticality()
            .mk_diag_base()
            .with_labels(
                locs.iter()
                    .map(|span| Label::primary(file_id.id, span.clone()))
                    .collect(),
            )
            .with_message(match self {
                DiagnosticKind::InvalidToken => {
                    if locs.len() > 1 {
                        "invalid tokens"
                    } else {
                        "invalid token"
                    }
                }
            })
    }
}

pub struct DiagnosticReporter<'a> {
    files: SimpleFiles<&'a str, &'a str>,
    errs: HashMap<FileIdent<'a>, HashMap<DiagnosticKind, Vec<Span>>>,
}

impl<'a> DiagnosticReporter<'a> {
    pub fn new() -> Self {
        Self {
            files: SimpleFiles::new(),
            errs: HashMap::new(),
        }
    }

    pub fn register_file(&mut self, name: &'a str, content: &'a str) -> FileIdent<'a> {
        FileIdent {
            id: self.files.add(name, content),
            name,
        }
    }

    pub fn report_diagnostic(&mut self, file_id: FileIdent<'a>, loc: Span, kind: DiagnosticKind) {
        self.errs
            .entry(file_id)
            .or_insert_with(|| HashMap::with_capacity(1));

        if !self.errs[&file_id].contains_key(&kind) {
            self.errs.get_mut(&file_id).unwrap().insert(kind, vec![loc]);
        } else {
            self.errs
                .get_mut(&file_id)
                .unwrap()
                .get_mut(&kind)
                .unwrap()
                .push(loc)
        }

        match kind.criticality() {
            DiagCriticality::Bug | DiagCriticality::CriticalImmediate => self.finish_and_exit(),
            DiagCriticality::CriticalFinishStep
            | DiagCriticality::Warn
            | DiagCriticality::Info
            | DiagCriticality::Help => {}
        }
    }

    pub fn fail_step_if_failed(&self) {
        for (_, diags) in self.errs.iter() {
            for kind in diags.keys() {
                if matches!(
                    kind.criticality(),
                    DiagCriticality::Bug
                        | DiagCriticality::CriticalImmediate
                        | DiagCriticality::CriticalFinishStep
                ) {
                    self.finish_and_exit()
                }
            }
        }
    }

    fn finish_and_exit(&self) -> ! {
        let writer = StandardStream::stderr(ColorChoice::Always);
        let config = term::Config::default();

        for (file_id, diags) in self.errs.iter() {
            for (kind, locs) in diags {
                let writer = &mut writer.lock();
                term::emit(writer, &config, &self.files, &kind.into_diag(locs, file_id))
                    .expect("can't even report errors :c");
            }
        }

        process::exit(1);
    }
}
